"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = require("path");
const fs_1 = require("fs");
const url_1 = require("url");
const graphql_config_1 = require("graphql-config");
const graphql_language_service_types_1 = require("graphql-language-service-types");
const graphql_language_service_interface_1 = require("graphql-language-service-interface");
const graphql_language_service_utils_1 = require("graphql-language-service-utils");
const GraphQLCache_1 = require("./GraphQLCache");
const languages_1 = require("./languages");
class MessageProcessor {
    constructor(logger) {
        this._textDocumentCache = new Map();
        this._isInitialized = false;
        this._willShutdown = false;
        this._logger = logger;
    }
    async handleInitializeRequest(params, _token, configDir) {
        if (!params) {
            throw new Error('`params` argument is required to initialize.');
        }
        const serverCapabilities = {
            capabilities: {
                completionProvider: { resolveProvider: true },
                definitionProvider: true,
                textDocumentSync: 1,
                hoverProvider: true,
            },
        };
        const rootPath = path_1.dirname(graphql_config_1.findGraphQLConfigFile(configDir ? configDir.trim() : params.rootPath));
        if (!rootPath) {
            throw new Error('`--configDir` option or `rootPath` argument is required.');
        }
        this._graphQLCache = await GraphQLCache_1.getGraphQLCache(rootPath);
        this._languageService = new graphql_language_service_interface_1.GraphQLLanguageService(this._graphQLCache);
        if (!serverCapabilities) {
            throw new Error('GraphQL Language Server is not initialized.');
        }
        this._isInitialized = true;
        this._logger.log(JSON.stringify({
            type: 'usage',
            messageType: 'initialize',
        }));
        return serverCapabilities;
    }
    async handleDidOpenOrSaveNotification(params) {
        if (!this._isInitialized) {
            return null;
        }
        if (!params || !params.textDocument) {
            throw new Error('`textDocument` argument is required.');
        }
        const { text, textDocument } = params;
        const { uri } = textDocument;
        const diagnostics = [];
        let contents = [];
        if (text || text === '') {
            contents = getQueryAndRange(text, uri);
            this._invalidateCache(textDocument, uri, contents);
        }
        else {
            const cachedDocument = this._getCachedDocument(uri);
            if (cachedDocument) {
                contents = cachedDocument.contents;
            }
        }
        await Promise.all(contents.map(async ({ query, range }) => {
            const results = await this._languageService.getDiagnostics(query, uri, this._isRelayCompatMode(query) ? false : true);
            if (results && results.length > 0) {
                diagnostics.push(...processDiagnosticsMessage(results, query, range));
            }
        }));
        this._logger.log(JSON.stringify({
            type: 'usage',
            messageType: 'textDocument/didOpen',
            projectName: this._graphQLCache
                .getGraphQLConfig()
                .getProjectNameForFile(uri),
            fileName: uri,
        }));
        return { uri, diagnostics };
    }
    async handleDidChangeNotification(params) {
        if (!this._isInitialized) {
            return null;
        }
        if (!params ||
            !params.textDocument ||
            !params.contentChanges ||
            !params.textDocument.uri) {
            throw new Error('`textDocument`, `textDocument.uri`, and `contentChanges` arguments are required.');
        }
        const textDocument = params.textDocument;
        const contentChanges = params.contentChanges;
        const contentChange = contentChanges[contentChanges.length - 1];
        const uri = textDocument.uri;
        const contents = getQueryAndRange(contentChange.text, uri);
        this._invalidateCache(textDocument, uri, contents);
        const cachedDocument = this._getCachedDocument(uri);
        if (!cachedDocument) {
            return null;
        }
        this._updateFragmentDefinition(uri, contents);
        this._updateObjectTypeDefinition(uri, contents);
        const diagnostics = [];
        await Promise.all(contents.map(async ({ query, range }) => {
            const results = await this._languageService.getDiagnostics(query, uri);
            if (results && results.length > 0) {
                diagnostics.push(...processDiagnosticsMessage(results, query, range));
            }
        }));
        this._logger.log(JSON.stringify({
            type: 'usage',
            messageType: 'textDocument/didChange',
            projectName: this._graphQLCache
                .getGraphQLConfig()
                .getProjectNameForFile(uri),
            fileName: uri,
        }));
        return { uri, diagnostics };
    }
    handleDidCloseNotification(params) {
        if (!this._isInitialized) {
            return;
        }
        if (!params || !params.textDocument) {
            throw new Error('`textDocument` is required.');
        }
        const textDocument = params.textDocument;
        const uri = textDocument.uri;
        if (this._textDocumentCache.has(uri)) {
            this._textDocumentCache.delete(uri);
        }
        this._logger.log(JSON.stringify({
            type: 'usage',
            messageType: 'textDocument/didClose',
            projectName: this._graphQLCache
                .getGraphQLConfig()
                .getProjectNameForFile(uri),
            fileName: uri,
        }));
    }
    handleShutdownRequest() {
        this._willShutdown = true;
        return;
    }
    handleExitNotification() {
        process.exit(this._willShutdown ? 0 : 1);
    }
    validateDocumentAndPosition(params) {
        if (!params ||
            !params.textDocument ||
            !params.textDocument.uri ||
            !params.position) {
            throw new Error('`textDocument`, `textDocument.uri`, and `position` arguments are required.');
        }
    }
    async handleCompletionRequest(params) {
        if (!this._isInitialized) {
            return [];
        }
        this.validateDocumentAndPosition(params);
        const textDocument = params.textDocument;
        const position = params.position;
        const cachedDocument = this._getCachedDocument(textDocument.uri);
        if (!cachedDocument) {
            throw new Error('A cached document cannot be found.');
        }
        const found = cachedDocument.contents.find(content => {
            const currentRange = content.range;
            if (currentRange && currentRange.containsPosition(position)) {
                return true;
            }
        });
        if (!found) {
            return [];
        }
        const { query, range } = found;
        if (range) {
            position.line -= range.start.line;
        }
        const result = await this._languageService.getAutocompleteSuggestions(query, position, textDocument.uri);
        this._logger.log(JSON.stringify({
            type: 'usage',
            messageType: 'textDocument/completion',
            projectName: this._graphQLCache
                .getGraphQLConfig()
                .getProjectNameForFile(textDocument.uri),
            fileName: textDocument.uri,
        }));
        return { items: result, isIncomplete: false };
    }
    async handleHoverRequest(params) {
        if (!this._isInitialized) {
            return { contents: [] };
        }
        this.validateDocumentAndPosition(params);
        const textDocument = params.textDocument;
        const position = params.position;
        const cachedDocument = this._getCachedDocument(textDocument.uri);
        if (!cachedDocument) {
            throw new Error('A cached document cannot be found.');
        }
        const found = cachedDocument.contents.find(content => {
            const currentRange = content.range;
            if (currentRange && currentRange.containsPosition(position)) {
                return true;
            }
        });
        if (!found) {
            return { contents: [] };
        }
        const { query, range } = found;
        if (range) {
            position.line -= range.start.line;
        }
        const result = await this._languageService.getHoverInformation(query, position, textDocument.uri);
        return {
            contents: result,
        };
    }
    async handleWatchedFilesChangedNotification(params) {
        if (!this._isInitialized) {
            return null;
        }
        return Promise.all(params.changes.map(async (change) => {
            if (change.type === graphql_language_service_types_1.FileChangeTypeKind.Created ||
                change.type === graphql_language_service_types_1.FileChangeTypeKind.Changed) {
                const uri = change.uri;
                const text = fs_1.readFileSync(new url_1.URL(uri).pathname).toString();
                const contents = getQueryAndRange(text, uri);
                this._updateFragmentDefinition(uri, contents);
                this._updateObjectTypeDefinition(uri, contents);
                const diagnostics = (await Promise.all(contents.map(async ({ query, range }) => {
                    const results = await this._languageService.getDiagnostics(query, uri);
                    if (results && results.length > 0) {
                        return processDiagnosticsMessage(results, query, range);
                    }
                    else {
                        return [];
                    }
                }))).reduce((left, right) => left.concat(right));
                this._logger.log(JSON.stringify({
                    type: 'usage',
                    messageType: 'workspace/didChangeWatchedFiles',
                    projectName: this._graphQLCache
                        .getGraphQLConfig()
                        .getProjectNameForFile(uri),
                    fileName: uri,
                }));
                return { uri, diagnostics };
            }
            else if (change.type === graphql_language_service_types_1.FileChangeTypeKind.Deleted) {
                this._graphQLCache.updateFragmentDefinitionCache(this._graphQLCache.getGraphQLConfig().configDir, change.uri, false);
                this._graphQLCache.updateObjectTypeDefinitionCache(this._graphQLCache.getGraphQLConfig().configDir, change.uri, false);
                return { uri: change.uri, diagnostics: [] };
            }
            return { uri: change.uri, diagnostics: [] };
        }));
    }
    async handleDefinitionRequest(params, _token) {
        if (!this._isInitialized) {
            return [];
        }
        if (!params || !params.textDocument || !params.position) {
            throw new Error('`textDocument` and `position` arguments are required.');
        }
        const textDocument = params.textDocument;
        const position = params.position;
        const cachedDocument = this._getCachedDocument(textDocument.uri);
        if (!cachedDocument) {
            throw new Error(`${textDocument.uri} is not available.`);
        }
        const found = cachedDocument.contents.find(content => {
            const currentRange = content.range;
            if (currentRange && currentRange.containsPosition(position)) {
                return true;
            }
        });
        if (!found) {
            return [];
        }
        const { query, range } = found;
        if (range) {
            position.line -= range.start.line;
        }
        const result = await this._languageService.getDefinition(query, position, textDocument.uri);
        const formatted = result
            ? result.definitions.map(res => {
                const defRange = res.range;
                return {
                    uri: res.path.indexOf('file://') === 0
                        ? res.path
                        : `file://${res.path}`,
                    range: defRange,
                };
            })
            : [];
        this._logger.log(JSON.stringify({
            type: 'usage',
            messageType: 'textDocument/definition',
            projectName: this._graphQLCache
                .getGraphQLConfig()
                .getProjectNameForFile(textDocument.uri),
            fileName: textDocument.uri,
        }));
        return formatted;
    }
    _isRelayCompatMode(query) {
        return (query.indexOf('RelayCompat') !== -1 ||
            query.indexOf('react-relay/compat') !== -1);
    }
    async _updateFragmentDefinition(uri, contents) {
        const rootDir = this._graphQLCache.getGraphQLConfig().configDir;
        await this._graphQLCache.updateFragmentDefinition(rootDir, new url_1.URL(uri).pathname, contents);
    }
    async _updateObjectTypeDefinition(uri, contents) {
        const rootDir = this._graphQLCache.getGraphQLConfig().configDir;
        await this._graphQLCache.updateObjectTypeDefinition(rootDir, new url_1.URL(uri).pathname, contents);
    }
    _getCachedDocument(uri) {
        if (this._textDocumentCache.has(uri)) {
            const cachedDocument = this._textDocumentCache.get(uri);
            if (cachedDocument) {
                return cachedDocument;
            }
        }
        return null;
    }
    _invalidateCache(textDocument, uri, contents) {
        if (this._textDocumentCache.has(uri)) {
            const cachedDocument = this._textDocumentCache.get(uri);
            if (cachedDocument &&
                textDocument.version &&
                cachedDocument.version < textDocument.version) {
                this._textDocumentCache.set(uri, {
                    version: textDocument.version,
                    contents,
                });
            }
        }
        else if (textDocument.version) {
            this._textDocumentCache.set(uri, {
                version: textDocument.version,
                contents,
            });
        }
    }
}
exports.MessageProcessor = MessageProcessor;
function getQueryAndRange(text, uri) {
    const languageConfig = languages_1.getLanguageFromExtension(path_1.extname(uri).slice(1));
    if (!languageConfig || !languageConfig.fileFilter(text)) {
        return [];
    }
    const templates = languageConfig.extractSources(text);
    return templates.map(({ template, range }) => ({ query: template, range }));
}
exports.getQueryAndRange = getQueryAndRange;
function processDiagnosticsMessage(results, query, range) {
    const queryLines = query.split('\n');
    const totalLines = queryLines.length;
    const lastLineLength = queryLines[totalLines - 1].length;
    const lastCharacterPosition = new graphql_language_service_utils_1.Position(totalLines, lastLineLength);
    const processedResults = results.filter(diagnostic => diagnostic.range.end.lessThanOrEqualTo(lastCharacterPosition));
    if (range) {
        const offset = range.start;
        return processedResults.map(diagnostic => (Object.assign(Object.assign({}, diagnostic), { range: new graphql_language_service_utils_1.Range(new graphql_language_service_utils_1.Position(diagnostic.range.start.line + offset.line, diagnostic.range.start.character), new graphql_language_service_utils_1.Position(diagnostic.range.end.line + offset.line, diagnostic.range.end.character)) })));
    }
    return processedResults;
}
//# sourceMappingURL=MessageProcessor.js.map